/* jl2005bcd_decompress.h
 *
 * Copyright (C) 2006-2010 Theodore Kilgore <kilgota@auburn.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301  USA
 */

#ifndef CAMLIBS_JL2005C_JL2005BCD_DECOMPRESS_H
#define CAMLIBS_JL2005C_JL2005BCD_DECOMPRESS_H


int
jl2005bcd_decompress(unsigned char *output, unsigned char *input,
					int inputsize, int get_thumbnail);

#endif /* !defined(CAMLIBS_JL2005C_JL2005BCD_DECOMPRESS_H) */

