/** \file
 *
 * \author Copyright 2001 Lutz Mueller <lutz@users.sf.net>
 *
 * \par License
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * \par
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * \par
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301  USA
 */

#ifndef LIBGPHOTO2_GPHOTO2_PORT_INFO_H
#define LIBGPHOTO2_GPHOTO2_PORT_INFO_H

/**
 * \internal GPPortInfo:
 *
 * The internals of this list are private.
 **/
struct _GPPortInfo {
	GPPortType type;	/**< \brief The type of this port. */
	char *name;		/**< \brief The name of this port (usb:) */
	char *path;		/**< \brief The path of this port (usb:001,023) */

	/* Private */
	char *library_filename;	/**< \brief Internal pathname of the port driver. Do not use outside of the port library. */
};

#endif /* !defined(LIBGPHOTO2_GPHOTO2_PORT_INFO_H) */
