/* $Id: crc16.h,v 1.2 2005/01/18 14:00:46 purbanec Exp $ */

/*

  Copyright (C) 2004 Peter Urbanec <toppy at urbanec.net>

  This file is part of puppy.

  puppy is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  puppy is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with puppy; if not, write to the
  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
  Boston, MA  02110-1301  USA

*/

#ifndef CAMLIBS_TOPFIELD_CRC16_H
#define CAMLIBS_TOPFIELD_CRC16_H

#include <sys/types.h>

unsigned short crc16_ansi(const void *data, size_t size);

#endif /* !defined(CAMLIBS_TOPFIELD_CRC16_H) */
